@@ -18,15 +18,17 @@
 from __future__ import division, print_function, absolute_import
 
 import math
-from decimal import Decimal
+import datetime as dt
+import operator
+from decimal import Decimal, InvalidOperation
 from numbers import Rational
 from fractions import Fraction
 
 from hypothesis.errors import InvalidArgument
 from hypothesis.control import assume
 from hypothesis.searchstrategy import SearchStrategy
-from hypothesis.internal.compat import ArgSpec, hrange, text_type, \
-    getargspec, integer_types, float_to_decimal, implements_iterator
+from hypothesis.internal.compat import hrange, text_type, integer_types, \
+    getfullargspec, implements_iterator
 from hypothesis.internal.floats import is_negative, float_to_int, \
     int_to_float, count_between_floats
 from hypothesis.utils.conventions import not_set
@@ -44,10 +46,12 @@ __all__ = [
     'tuples', 'lists', 'sets', 'frozensets', 'iterables',
     'dictionaries', 'fixed_dictionaries',
     'sampled_from', 'permutations',
+    'datetimes', 'dates', 'times', 'timedeltas',
     'builds',
     'randoms', 'random_module',
     'recursive', 'composite',
     'shared', 'runner', 'data',
+    'deferred',
 ]
 
 _strategies = set()
@@ -101,17 +105,19 @@ def cacheable(fn):
 
 
 def defines_strategy(strategy_definition):
-    from hypothesis.searchstrategy.deferred import DeferredStrategy
+    from hypothesis.searchstrategy.lazy import LazyStrategy
     _strategies.add(strategy_definition.__name__)
 
     @proxies(strategy_definition)
     def accept(*args, **kwargs):
-        return DeferredStrategy(strategy_definition, args, kwargs)
+        return LazyStrategy(strategy_definition, args, kwargs)
     return accept
 
 
 class Nothing(SearchStrategy):
-    is_empty = True
+    @property
+    def is_empty(self):
+        return True
 
     def do_draw(self, data):
         data.mark_invalid()
@@ -173,19 +179,8 @@ def one_of(*args):
             args = tuple(args[0])
         except TypeError:
             pass
-
-    strategies = []
-    for arg in args:
-        check_strategy(arg)
-        if not arg.is_empty:
-            strategies.extend([s for s in arg.branches if not s.is_empty])
-
-    if not strategies:
-        return nothing()
-    if len(strategies) == 1:
-        return strategies[0]
     from hypothesis.searchstrategy.strategies import OneOfStrategy
-    return OneOfStrategy(strategies)
+    return OneOfStrategy(args)
 
 
 @cacheable
@@ -310,27 +305,18 @@ def floats(
     elif min_value is not None and max_value is not None:
         if min_value == max_value:
             return just(min_value)
-        elif math.isinf(max_value - min_value):
-            assert min_value < 0 and max_value > 0
-            return floats(min_value=0, max_value=max_value) | floats(
-                min_value=min_value, max_value=0
-            )
+        elif is_negative(min_value):
+            if is_negative(max_value):
+                return floats(min_value=-max_value, max_value=-min_value).map(
+                    operator.neg
+                )
+            else:
+                return floats(min_value=0.0, max_value=max_value) | floats(
+                    min_value=0.0, max_value=-min_value).map(operator.neg)
         elif count_between_floats(min_value, max_value) > 1000:
             return FixedBoundedFloatStrategy(
                 lower_bound=min_value, upper_bound=max_value
             )
-        elif is_negative(max_value):
-            assert is_negative(min_value)
-            ub_int = float_to_int(max_value)
-            lb_int = float_to_int(min_value)
-            assert ub_int <= lb_int
-            return integers(min_value=ub_int, max_value=lb_int).map(
-                int_to_float
-            )
-        elif is_negative(min_value):
-            return floats(min_value=min_value, max_value=-0.0) | floats(
-                min_value=0, max_value=max_value
-            )
         else:
             ub_int = float_to_int(max_value)
             lb_int = float_to_int(min_value)
@@ -342,7 +328,7 @@ def floats(
         if min_value < 0:
             result = floats(
                 min_value=0.0
-            ) | floats(min_value=min_value, max_value=0.0)
+            ) | floats(min_value=min_value, max_value=-0.0)
         else:
             result = (
                 floats(allow_infinity=allow_infinity, allow_nan=False).map(
@@ -358,7 +344,7 @@ def floats(
             result = floats(
                 min_value=0.0,
                 max_value=max_value,
-            ) | floats(max_value=0.0)
+            ) | floats(max_value=-0.0)
         else:
             result = (
                 floats(allow_infinity=allow_infinity, allow_nan=False).map(
@@ -585,7 +571,7 @@ def fixed_dictionaries(mapping):
 
     """
     from hypothesis.searchstrategy.collections import FixedKeysDictStrategy
-    check_type(dict, mapping)
+    check_type(dict, mapping, 'mapping')
     for v in mapping.values():
         check_strategy(v)
     for v in mapping.values():
@@ -836,31 +822,85 @@ def fractions(min_value=None, max_value=None, max_denominator=None):
 
 @cacheable
 @defines_strategy
-def decimals(
-    min_value=None, max_value=None, allow_nan=None, allow_infinity=None
-):
-    """Generates instances of decimals.Decimal.
+def decimals(min_value=None, max_value=None,
+             allow_nan=None, allow_infinity=None, places=None):
+    """Generates instances of decimals.Decimal, which may be:
 
-    - If min_value is not None, all values will be >= min_value.
-    - If max_value is not None, all values will be <= max_value.
-    - If min_value or max_value is not None, it is an error to enable
-      allow_nan.
-    - If both min_value and max_value are not None, it is an error to enable
-      allow_infinity.
+    - A finite rational number, between ``min_value`` and ``max_value``.
+    - Not a Number, if ``allow_nan`` is True.  None means "allow NaN, unless
+      ``min__value`` and ``max_value`` are not None".
+    - Positive or negative infinity, if ``max_value`` and ``min_value``
+      respectively are None, and ``allow_infinity`` is not False.  None means
+      "allow infinity, unless excluded by the min and max values".
 
-    Where not explicitly ruled out by the bounds, all of infinity, -infinity
-    and NaN are possible values generated by this strategy.
+    Note that where floats have one `NaN` value, Decimals have four: signed,
+    and either *quiet* or *signalling*.  See `the decimal module docs
+    <https://docs.python.org/3/library/decimal.html#special-values>`_ for
+    more information on special values.
+
+    If ``places`` is not None, all finite values drawn from the strategy will
+    have that number of digits after the decimal place.
 
     """
-    floats_strategy = floats(
-        min_value=min_value, max_value=max_value,
-        allow_nan=allow_nan, allow_infinity=allow_infinity)
-    fractions_strategy = fractions(min_value=min_value, max_value=max_value)
+    # Convert min_value and max_value to Decimal values, and validate args
+    check_valid_integer(places)
+    if places is not None and places < 0:
+        raise InvalidArgument('places=%r may not be negative' % places)
 
-    return (
-        floats_strategy.map(float_to_decimal) |
-        fractions_strategy.map(lambda f: Decimal(f.numerator) / f.denominator)
-    )
+    if min_value is not None:
+        min_value = Decimal(min_value)
+        if min_value.is_infinite() and min_value < 0:
+            if not (allow_infinity or allow_infinity is None):
+                raise InvalidArgument('allow_infinity=%r, but min_value=%r'
+                                      % (allow_infinity, min_value))
+            min_value = None
+        elif not min_value.is_finite():
+            # This could be positive infinity, quiet NaN, or signalling NaN
+            raise InvalidArgument(u'Invalid min_value=%r' % min_value)
+    if max_value is not None:
+        max_value = Decimal(max_value)
+        if max_value.is_infinite() and max_value > 0:
+            if not (allow_infinity or allow_infinity is None):
+                raise InvalidArgument('allow_infinity=%r, but max_value=%r'
+                                      % (allow_infinity, max_value))
+            max_value = None
+        elif not max_value.is_finite():
+            raise InvalidArgument(u'Invalid max_value=%r' % max_value)
+    check_valid_bound(min_value, 'min_value')
+    check_valid_bound(max_value, 'max_value')
+    check_valid_interval(min_value, max_value, 'min_value', 'max_value')
+
+    if allow_infinity and (None not in (min_value, max_value)):
+        raise InvalidArgument('Cannot allow infinity between finite bounds')
+    # Set up a strategy for finite decimals
+    if places is not None:
+        # Fixed-point decimals are basically integers with a scale factor
+
+        def try_quantize(d):
+            try:
+                return d.quantize(factor)
+            except InvalidOperation:  # pragma: no cover
+                return None
+        factor = Decimal(10) ** -places
+        max_num = max_value / factor if max_value is not None else None
+        min_num = min_value / factor if min_value is not None else None
+        strat = integers(min_value=min_num, max_value=max_num)\
+            .map(lambda d: try_quantize(d * factor))\
+            .filter(lambda d: d is not None)
+    else:
+        # Otherwise, they're like fractions featuring a power of ten
+        strat = fractions(
+            min_value=min_value, max_value=max_value
+        ).map(lambda f: Decimal(f.numerator) / f.denominator)
+    # Compose with sampled_from for infinities and NaNs as appropriate
+    special = []
+    if allow_nan or (allow_nan is None and (None in (min_value, max_value))):
+        special.extend(map(Decimal, ('NaN', '-NaN', 'sNaN', '-sNaN')))
+    if allow_infinity or (allow_infinity is max_value is None):
+        special.append(Decimal('Infinity'))
+    if allow_infinity or (allow_infinity is min_value is None):
+        special.append(Decimal('-Infinity'))
+    return strat | sampled_from(special)
 
 
 def recursive(base, extend, max_leaves=100):
@@ -910,18 +950,122 @@ def permutations(values):
     return PermutationStrategy()
 
 
+@defines_strategy
+def datetimes(min_datetime=dt.datetime.min, max_datetime=dt.datetime.max,
+              timezones=none()):
+    """A strategy for generating datetimes, which may be timezone-aware.
+
+    This strategy works by drawing a naive datetime between ``min_datetime``
+    and ``max_datetime``, which must both be naive (have no timezone).
+
+    ``timezones`` must be a strategy that generates tzinfo objects (or None,
+    which is valid for naive datetimes).  A value drawn from this strategy
+    will be added to a naive datetime, and the resulting tz-aware datetime
+    returned.
+
+    .. note::
+        tz-aware datetimes from this strategy may be ambiguous or non-existent
+        due to daylight savings, leap seconds, timezone and calendar
+        adjustments, etc.  This is intentional, as malformed timestamps are a
+        common source of bugs.
+
+    :py:func:`hypothesis.extra.timezones` requires the ``pytz`` package, but
+    provides all timezones in the Olsen database.  If you also want to allow
+    naive datetimes, combine strategies like ``none() | timezones()``.
+
+    Alternatively, you can create a list of the timezones you wish to allow
+    (e.g. from the standard library, ``datetutil``, or ``pytz``) and use
+    :py:func:`sampled_from`.  Ensure that simple values such as None or UTC
+    are at the beginning of the list for proper minimisation.
+
+    """
+    # Why must bounds be naive?  In principle, we could also write a strategy
+    # that took aware bounds, but the API and validation is much harder.
+    # If you want to generate datetimes between two particular momements in
+    # time I suggest (a) just filtering out-of-bounds values; (b) if bounds
+    # are very close, draw a value and subtract it's UTC offset, handling
+    # overflows and nonexistent times; or (c) do something customised to
+    # handle datetimes in e.g. a four-microsecond span which is not
+    # representable in UTC.  Handling (d), all of the above, leads to a much
+    # more complex API for all users and a useful feature for very few.
+    from hypothesis.searchstrategy.datetime import DatetimeStrategy
+
+    check_type(dt.datetime, min_datetime, 'min_datetime')
+    check_type(dt.datetime, max_datetime, 'max_datetime')
+    if min_datetime.tzinfo is not None:
+        raise InvalidArgument('min_datetime=%r must not have tzinfo'
+                              % (min_datetime,))
+    if max_datetime.tzinfo is not None:
+        raise InvalidArgument('max_datetime=%r must not have tzinfo'
+                              % (max_datetime,))
+    check_valid_interval(min_datetime, max_datetime,
+                         'min_datetime', 'max_datetime')
+    if not isinstance(timezones, SearchStrategy):
+        raise InvalidArgument(
+            'timezones=%r must be a SearchStrategy that can provide tzinfo '
+            'for datetimes (either None or dt.tzinfo objects)' % (timezones,))
+    return DatetimeStrategy(min_datetime, max_datetime, timezones)
+
+
+@defines_strategy
+def dates(min_date=dt.date.min, max_date=dt.date.max):
+    """A strategy for dates between ``min_date`` and ``max_date``."""
+    from hypothesis.searchstrategy.datetime import DateStrategy
+
+    check_type(dt.date, min_date, 'min_date')
+    check_type(dt.date, max_date, 'max_date')
+    check_valid_interval(min_date, max_date, 'min_date', 'max_date')
+    if min_date == max_date:
+        return just(min_date)
+    return DateStrategy(min_date, max_date)
+
+
+@defines_strategy
+def times(min_time=dt.time.min, max_time=dt.time.max, timezones=none()):
+    """A strategy for times between ``min_time`` and ``max_time``.
+
+    The ``timezones`` argument is handled as for :py:func:`datetimes`.
+
+    """
+    check_type(dt.time, min_time, 'min_time')
+    check_type(dt.time, max_time, 'max_time')
+    if min_time.tzinfo is not None:
+        raise InvalidArgument('min_time=%r must not have tzinfo' % min_time)
+    if max_time.tzinfo is not None:
+        raise InvalidArgument('max_time=%r must not have tzinfo' % max_time)
+    check_valid_interval(min_time, max_time, 'min_time', 'max_time')
+    day = dt.date(2000, 1, 1)
+    return datetimes(min_datetime=dt.datetime.combine(day, min_time),
+                     max_datetime=dt.datetime.combine(day, max_time),
+                     timezones=timezones).map(lambda t: t.timetz())
+
+
+@defines_strategy
+def timedeltas(min_delta=dt.timedelta.min, max_delta=dt.timedelta.max):
+    """A strategy for timedeltas between ``min_delta`` and ``max_delta``."""
+    from hypothesis.searchstrategy.datetime import TimedeltaStrategy
+
+    check_type(dt.timedelta, min_delta, 'min_delta')
+    check_type(dt.timedelta, max_delta, 'max_delta')
+    check_valid_interval(min_delta, max_delta, 'min_delta', 'max_delta')
+    if min_delta == max_delta:
+        return just(min_delta)
+    return TimedeltaStrategy(min_delta=min_delta, max_delta=max_delta)
+
+
 @cacheable
 def composite(f):
     """Defines a strategy that is built out of potentially arbitrarily many
     other strategies.
 
-    This is intended to be used as a decorator. See the full
-    documentation for more details about how to use this function.
+    This is intended to be used as a decorator. See
+    :ref:`the full documentation for more details <composite-strategies>`
+    about how to use this function.
 
     """
 
     from hypothesis.internal.reflection import define_function_signature
-    argspec = getargspec(f)
+    argspec = getfullargspec(f)
 
     if (
         argspec.defaults is not None and
@@ -935,10 +1079,9 @@ def composite(f):
             'positional argument.'
         )
 
-    new_argspec = ArgSpec(
-        args=argspec.args[1:], varargs=argspec.varargs,
-        keywords=argspec.keywords, defaults=argspec.defaults
-    )
+    annots = {k: v for k, v in argspec.annotations.items()
+              if k in (argspec.args + argspec.kwonlyargs + ['return'])}
+    new_argspec = argspec._replace(args=argspec.args[1:], annotations=annots)
 
     @defines_strategy
     @define_function_signature(f.__name__, f.__doc__, new_argspec)
@@ -1069,7 +1212,8 @@ def data():
     """This isn't really a normal strategy, but instead gives you an object
     which can be used to draw data interactively from other strategies.
 
-    It can only be used within @given, not find. This is because the lifetime
+    It can only be used within :func:`@given <hypothesis.given>`, not
+    :func:`find() <hypothesis.find>`. This is because the lifetime
     of the object cannot outlast the test body.
 
     See :ref:`the rest of the documentation <interactive-draw>` for more
@@ -1130,15 +1274,17 @@ def data():
 # Private API below here
 
 
-def check_type(typ, arg):
+def check_type(typ, arg, name=''):
+    if name:
+        name += '='
     if not isinstance(arg, typ):
         if isinstance(typ, type):
             typ_string = typ.__name__
         else:
             typ_string = 'one of %s' % (
                 ', '.join(t.__name__ for t in typ))
-        raise InvalidArgument(
-            'Expected %s but got %r' % (typ_string, arg,))
+        raise InvalidArgument('Expected %s but got %s%r (type=%s)'
+                              % (typ_string, name, arg, type(arg).__name__))
 
 
 def check_strategy(arg):
@@ -1220,4 +1366,38 @@ def check_valid_sizes(min_size, average_size, max_size):
 
 
 _AVERAGE_LIST_LENGTH = 5.0
+
+
+def deferred(definition):
+    """A deferred strategy allows you to write a strategy that references other
+    strategies that have not yet been defined. This allows for the easy
+    definition of recursive and mutually recursive strategies.
+
+    The definition argument should be a zero-argument function that returns a
+    strategy. It will be evaluated the first time the strategy is used to
+    produce an example.
+
+    Example usage:
+
+    >>> import hypothesis.strategies as st
+    >>> x = st.deferred(lambda: st.booleans() | st.tuples(x, x))
+    >>> x.example()
+    (False, (False, True))
+    >>> x.example()
+    True
+
+    Mutual recursion also works fine:
+
+    >>> a = st.deferred(lambda: st.booleans() | b)
+    >>> b = st.deferred(lambda: st.tuples(a, a))
+    >>> a.example()
+    (((True, True), False), True)
+    >>> b.example()
+    (((( False, ((True, False), (True, True))), True), False), True)
+
+    """
+    from hypothesis.searchstrategy.deferred import DeferredStrategy
+    return DeferredStrategy(definition)
+
+
 assert _strategies.issubset(set(__all__)), _strategies - set(__all__)
